@@ -1,18 +1,31 @@
 import os
-import shutil
+import subprocess
 
 from vcstool.executor import USE_COLOR
 
 from .vcs_base import VcsClientBase
 from .vcs_base import which
+from ..util import rmtree
 
 
 class GitClient(VcsClientBase):
 
     type = 'git'
     _executable = None
+    _git_version = None
     _config_color_is_auto = None
 
+    @classmethod
+    def get_git_version(cls):
+        if cls._git_version is None:
+            output = subprocess.check_output(['git', '--version'])
+            prefix = b'git version '
+            assert output.startswith(prefix)
+            output = output[len(prefix):].rstrip()
+            cls._git_version = [
+                int(x) for x in output.split(b'.') if x != b'windows']
+        return cls._git_version
+
     @staticmethod
     def is_repository(path):
         return os.path.isdir(os.path.join(path, '.git'))
@@ -28,7 +41,7 @@ class GitClient(VcsClientBase):
         if not command.all and not result['returncode']:
             # only show current branch
             lines = result['output'].splitlines()
-            lines = [l[2:] for l in lines if l.startswith('* ')]
+            lines = [line[2:] for line in lines if line.startswith('* ')]
             result['output'] = '\n'.join(lines)
 
         return result
@@ -146,48 +159,51 @@ class GitClient(VcsClientBase):
                 if ref not in refs:
                     continue
 
-                # check if there is exactly one tag pointing to that ref
-                cmd_tags = [
-                    GitClient._executable, 'tag', '--points-at', ref]
-                result_tags = self._run_command(cmd_tags)
-                if result_tags['returncode']:
-                    result_tags['output'] = \
-                        "Could not determine tags for ref '%s': " % \
-                        ref + result_tags['output']
-                    return result_tags
-                tags = result_tags['output'].splitlines()
-                tag_or_ref = ref
-                if len(tags) == 1:
-                    tag = tags[0]
-                    # double check that the tag is part of the remote
-                    # and references the same hash
-                    cmd_ls_remote = [
-                        GitClient._executable, 'ls-remote', remote,
-                        'refs/tags/' + tag]
-                    result_ls_remote = self._run_command(cmd_ls_remote)
-                    if result_ls_remote['returncode']:
-                        result_ls_remote['output'] = \
-                            "Could not check remote tags for '%s': " % \
-                            remote + result_ls_remote['output']
-                        return result_ls_remote
-                    matches = result_ls_remote['output'].splitlines()
-                    if len(matches) == 1 and matches[0].split()[0] == ref:
-                        tag_or_ref = tag
+                cmds = [result_ref['cmd']]
+                if command.with_tags:
+                    # check if there is exactly one tag pointing to that ref
+                    cmd_tags = [
+                        GitClient._executable, 'tag', '--points-at', ref]
+                    result_tags = self._run_command(cmd_tags)
+                    if result_tags['returncode']:
+                        result_tags['output'] = \
+                            "Could not determine tags for ref '%s': " % \
+                            ref + result_tags['output']
+                        return result_tags
+                    cmds.append(result_tags['cmd'])
+                    tags = result_tags['output'].splitlines()
+                    if len(tags) == 1:
+                        tag = tags[0]
+                        # double check that the tag is part of the remote
+                        # and references the same hash
+                        cmd_ls_remote = [
+                            GitClient._executable, 'ls-remote', remote,
+                            'refs/tags/' + tag]
+                        result_ls_remote = self._run_command(cmd_ls_remote)
+                        if result_ls_remote['returncode']:
+                            result_ls_remote['output'] = \
+                                "Could not check remote tags for '%s': " % \
+                                remote + result_ls_remote['output']
+                            return result_ls_remote
+                        matches = self._get_hash_ref_tuples(
+                            result_ls_remote['output'])
+                        if len(matches) == 1 and matches[0][0] == ref:
+                            ref = tag
 
                 # determine url of remote
                 result_url = self._get_remote_url(remote)
                 if result_url['returncode']:
                     return result_url
                 url = result_url['output']
-                # the result is the remote url and the hash
+                cmds.append(result_url['cmd'])
+
+                # the result is the remote url and the hash/tag
                 return {
-                    'cmd': ' && '.join([
-                        result_ref['cmd'], result_tags['cmd'],
-                        result_url['cmd']]),
+                    'cmd': ' && '.join(cmds),
                     'cwd': self.path,
-                    'output': '\n'.join([url, tag_or_ref]),
+                    'output': '\n'.join([url, ref]),
                     'returncode': 0,
-                    'export_data': {'url': url, 'version': tag_or_ref}
+                    'export_data': {'url': url, 'version': ref}
                 }
 
             return {
@@ -225,6 +241,14 @@ class GitClient(VcsClientBase):
                 if url == command.url:
                     break
             else:
+                if command.skip_existing:
+                    return {
+                        'cmd': '',
+                        'cwd': self.path,
+                        'output':
+                            'Skipped existing repository with different URL',
+                        'returncode': 0
+                    }
                 if not command.force:
                     return {
                         'cmd': '',
@@ -235,13 +259,22 @@ class GitClient(VcsClientBase):
                         'returncode': 1
                     }
                 try:
-                    shutil.rmtree(self.path)
+                    rmtree(self.path)
                 except OSError:
                     os.remove(self.path)
+
+        elif command.skip_existing and os.path.exists(self.path):
+            return {
+                'cmd': '',
+                'cwd': self.path,
+                'output': 'Skipped existing directory',
+                'returncode': 0
+            }
+
         elif command.force and os.path.exists(self.path):
             # Not empty, not a git repository
             try:
-                shutil.rmtree(self.path)
+                rmtree(self.path)
             except OSError:
                 os.remove(self.path)
 
@@ -250,9 +283,11 @@ class GitClient(VcsClientBase):
             return not_exist
 
         if GitClient.is_repository(self.path):
-            if command.version:
+            if command.skip_existing:
+                checkout_version = None
+            elif command.version:
                 checkout_version = command.version
-            elif not command.skip_existing:
+            else:
                 # determine remote HEAD branch
                 cmd_remote = [GitClient._executable, 'remote', 'show', remote]
                 # override locale in order to parse output
@@ -275,34 +310,126 @@ class GitClient(VcsClientBase):
                         'Could not determine remote HEAD branch of ' \
                         "repository '%s': %s" % (url, result_remote['output'])
                     return result_remote
-            else:
-                checkout_version = None
 
             # fetch updates for existing repo
             cmd_fetch = [GitClient._executable, 'fetch', remote]
+            if command.shallow:
+                result_version_type = self._check_version_type(
+                    command.url, checkout_version)
+                if result_version_type['returncode']:
+                    return result_version_type
+                version_type = result_version_type['version_type']
+                if version_type == 'branch':
+                    cmd_fetch.append(
+                        'refs/heads/%s:refs/remotes/%s/%s' %
+                        (checkout_version, remote, checkout_version))
+                elif version_type == 'hash':
+                    cmd_fetch.append(checkout_version)
+                elif version_type == 'tag':
+                    cmd_fetch.append(
+                        '+refs/tags/%s:refs/tags/%s' %
+                        (checkout_version, checkout_version))
+                else:
+                    assert False
+                cmd_fetch += ['--depth', '1']
             result_fetch = self._run_command(cmd_fetch, retry=command.retry)
             if result_fetch['returncode']:
                 return result_fetch
             cmd = result_fetch['cmd']
             output = result_fetch['output']
 
-        else:
-            cmd_clone = [GitClient._executable, 'clone', command.url, '.']
-            result_clone = self._run_command(cmd_clone, retry=command.retry)
-            if result_clone['returncode']:
-                result_clone['output'] = \
-                    "Could not clone repository '%s': %s" % \
-                    (command.url, result_clone['output'])
-                return result_clone
-            cmd = result_clone['cmd']
-            output = result_clone['output']
+            # ensure that a tracking branch exists which can be checked out
+            if command.shallow and version_type == 'branch':
+                cmd_show_ref = [
+                    GitClient._executable, 'show-ref',
+                    'refs/heads/%s' % checkout_version]
+                result_show_ref = self._run_command(cmd_show_ref)
+                if result_show_ref['returncode']:
+                    # creating tracking branch
+                    cmd_branch = [
+                        GitClient._executable, 'branch', checkout_version,
+                        '%s/%s' % (remote, checkout_version)]
+                    result_branch = self._run_command(cmd_branch)
+                    if result_branch['returncode']:
+                        result_branch['output'] = \
+                            "Could not create branch '%s': %s" % \
+                            (checkout_version, result_branch['output'])
+                        return result_branch
+                    cmd += ' && ' + ' '.join(cmd_branch)
+                    output = '\n'.join([output, result_branch['output']])
 
-            checkout_version = command.version
+        else:
+            version_type = None
+            if command.version:
+                result_version_type = self._check_version_type(
+                    command.url, command.version)
+                if result_version_type['returncode']:
+                    return result_version_type
+                version_type = result_version_type['version_type']
+
+            if not command.shallow or version_type in (None, 'branch'):
+                cmd_clone = [GitClient._executable, 'clone', command.url, '.']
+                if version_type == 'branch':
+                    cmd_clone += ['-b', command.version]
+                    checkout_version = None
+                else:
+                    checkout_version = command.version
+                if command.shallow:
+                    cmd_clone += ['--depth', '1']
+                result_clone = self._run_command(
+                    cmd_clone, retry=command.retry)
+                if result_clone['returncode']:
+                    result_clone['output'] = \
+                        "Could not clone repository '%s': %s" % \
+                        (command.url, result_clone['output'])
+                    return result_clone
+                cmd = result_clone['cmd']
+                output = result_clone['output']
+            else:
+                # getting a hash or tag with a depth of 1 can't use 'clone'
+                cmd_init = [GitClient._executable, 'init']
+                result_init = self._run_command(cmd_init)
+                if result_init['returncode']:
+                    return result_init
+                cmd = result_init['cmd']
+                output = result_init['output']
+
+                cmd_remote_add = [
+                    GitClient._executable, 'remote', 'add', 'origin',
+                    command.url]
+                result_remote_add = self._run_command(cmd_remote_add)
+                if result_remote_add['returncode']:
+                    return result_remote_add
+                cmd += ' && ' + ' '.join(cmd_remote_add)
+                output = '\n'.join([output, result_remote_add['output']])
+
+                cmd_fetch = [GitClient._executable, 'fetch', 'origin']
+                if version_type == 'hash':
+                    cmd_fetch.append(command.version)
+                elif version_type == 'tag':
+                    cmd_fetch.append(
+                        'refs/tags/%s:refs/tags/%s' %
+                        (command.version, command.version))
+                else:
+                    assert False
+                cmd_fetch += ['--depth', '1']
+                result_fetch = self._run_command(
+                    cmd_fetch, retry=command.retry)
+                if result_fetch['returncode']:
+                    return result_fetch
+                cmd += ' && ' + ' '.join(cmd_fetch)
+                output = '\n'.join([output, result_fetch['output']])
+
+                checkout_version = command.version
 
         if checkout_version:
             cmd_checkout = [
-                GitClient._executable, 'checkout', checkout_version]
+                GitClient._executable, 'checkout', checkout_version, '--']
             result_checkout = self._run_command(cmd_checkout)
+            if result_checkout['returncode']:
+                if self.get_git_version() < [1, 8, 4, 3]:
+                    cmd_checkout.pop()
+                    result_checkout = self._run_command(cmd_checkout)
             if result_checkout['returncode']:
                 result_checkout['output'] = \
                     "Could not checkout ref '%s': %s" % \
@@ -352,6 +479,37 @@ class GitClient(VcsClientBase):
             'returncode': 0 if remote_urls else 1
         }
 
+    def _check_version_type(self, url, version):
+        cmd = [GitClient._executable, 'ls-remote', url, version]
+        result = self._run_command(cmd)
+        if result['returncode']:
+            result['output'] = 'Could not determine ref type of version: ' + \
+                result['output']
+            return result
+        if not result['output']:
+            result['version_type'] = 'hash'
+            return result
+
+        refs = {}
+        for hash_, ref in self._get_hash_ref_tuples(result['output']):
+            refs[ref] = hash_
+
+        tag_ref = 'refs/tags/' + version
+        branch_ref = 'refs/heads/' + version
+        if tag_ref in refs and branch_ref in refs:
+            if refs[tag_ref] != refs[branch_ref]:
+                result['returncode'] = 1
+                result['output'] = 'The version ref is a branch as well as ' \
+                    'tag but with different hashes'
+                return result
+        if tag_ref in refs:
+            result['version_type'] = 'tag'
+        elif branch_ref in refs:
+            result['version_type'] = 'branch'
+        else:
+            result['version_type'] = 'hash'
+        return result
+
     def log(self, command):
         self._check_executable()
         if command.limit_tag:
@@ -381,8 +539,11 @@ class GitClient(VcsClientBase):
             cmd = [GitClient._executable, 'log', '%s..' % result_tag['output']]
         else:
             cmd = [GitClient._executable, 'log']
+        cmd += ['--decorate']
         if command.limit != 0:
             cmd += ['-%d' % command.limit]
+        if not command.verbose:
+            cmd += ['--pretty=short']
         self._check_color(cmd)
         return self._run_command(cmd)
 
@@ -390,7 +551,29 @@ class GitClient(VcsClientBase):
         self._check_executable()
         cmd = [GitClient._executable, 'pull']
         self._check_color(cmd)
-        return self._run_command(cmd)
+        result = self._run_command(cmd)
+
+        if result['returncode']:
+            # check for detached HEAD
+            cmd_rev_parse = [
+                GitClient._executable, 'rev-parse', '--abbrev-ref', 'HEAD']
+            result_rev_parse = self._run_command(cmd_rev_parse)
+            if result_rev_parse['returncode']:
+                result_rev_parse['output'] = 'Could not determine ref: ' + \
+                    result_rev_parse['output']
+                return result_rev_parse
+            detached = result_rev_parse['output'] == 'HEAD'
+
+            if detached:
+                # warn but not fail about the inability to pull a detached head
+                return {
+                    'cmd': '',
+                    'cwd': self.path,
+                    'output': result['output'],
+                    'returncode': 0,
+                }
+
+        return result
 
     def push(self, _command):
         self._check_executable()
@@ -459,10 +642,10 @@ class GitClient(VcsClientBase):
         if command.version:
             hashes = []
             refs = []
-            output_lines = result_ls_remote['output'].splitlines()
 
-            for line in output_lines:
-                hash_and_ref = line.split()
+            for hash_and_ref in self._get_hash_ref_tuples(
+                result_ls_remote['output']
+            ):
                 hashes.append(hash_and_ref[0])
 
                 # ignore pull request refs
@@ -528,6 +711,18 @@ class GitClient(VcsClientBase):
         assert GitClient._executable is not None, \
             "Could not find 'git' executable"
 
+    def _get_hash_ref_tuples(self, ls_remote_output):
+        tuples = []
+        for line in ls_remote_output.splitlines():
+            if line.startswith('#'):
+                continue
+            try:
+                hash_, ref = line.split(None, 1)
+            except ValueError:
+                continue
+            tuples.append((hash_, ref))
+        return tuples
+
 
 if not GitClient._executable:
     GitClient._executable = which('git')
